#ifndef ARR_ELEMENT_H
#define ARR_ELEMENT_H

#include "pocofwd.h"
#include <stdlib.h>
#include <assert.h>

/*
 * Interface for container elements contains pointers to functions
 * that: assign a new value to an element, compare two elements,
 * construct a new element with a default value, copy an element, and
 * destroy an element.
 *
 * refcount variable counts references to this object.  ElementInterface_copy
 * increments refcount; ElementInterface_destroy decrements it.  The
 * memory is freed when refcount drops to zero.
 */
struct ElementInterface {
    void *param;  /* Passed to constructor; not locally owned. */
    void *(*assign) (void *lhs, const void *rhs);
    int   (*comparepp) (const void *lhs, const void *rhs);
    void *(*construct) (void *param);
    void *(*copy) (const void *rhs);
    void  (*destroy) (void *obj);
    int   refcount;  
};

/** Prototypes **/
inline static ElementInterface *
ElementInterface_construct(void *(*assign)(void *lhs, const void *rhs),
			   int  (*comparepp)(const void *lhs,
					     const void *rhs),
			   void *(*construct) (void *param),
			   void *(*copy) (const void *rhs),
			   void  (*destroy) (void *obj),
			   void *param
			   );
inline static void ElementInterface_destroy(ElementInterface *eif);
inline static ElementInterface *ElementInterface_copy(ElementInterface *old);
inline static int ElementInterface_comparepp(const ElementInterface *eif,
					   const void *l, const void *r );
inline static int ElementInterface_compare(const ElementInterface *eif,
					   const void *l, const void *r );


/** Default constructor **/
inline static ElementInterface *
ElementInterface_construct(void *(*assign)(void *lhs, const void *rhs),
			   int   (*comparepp) (const void *lhs,
					     const void *rhs),
			   void *(*construct) (void * param),
			   void *(*copy) (const void *rhs),
			   void  (*destroy) (void *obj),
			   void *param
			   ) {
    ElementInterface * eif
	= (ElementInterface *) malloc(sizeof(ElementInterface));
    if(eif == NULL)
	return NULL;
    eif->assign = assign;
    eif->comparepp = comparepp;
    eif->construct = construct;
    eif->copy = copy;
    eif->destroy = destroy;
    eif->param = param; /* not locally owned */
    eif->refcount = 1;

    return eif;
}

/** Copy constructor increments refcount and returns its argument **/
inline static ElementInterface *
ElementInterface_copy(ElementInterface *old) {
    assert(old);
    ++old->refcount;
    return old;
}

/**
 * Destroy operator decrements refcount.  When refcount drops to
 * zero, memory is released.
 **/			   
inline static void 
ElementInterface_destroy(ElementInterface *eif){
    --eif->refcount;
    if(eif->refcount == 0)
	free(eif);
}

/** Compare pointers to pointers to two elements **/
inline static int 
ElementInterface_comparepp(const ElementInterface *eif,
			 const void *l, const void *r ) {
    return eif->comparepp(l, r);
}

/** Compare pointers to two elements **/
inline static int 
ElementInterface_compare(const ElementInterface *eif,
			 const void *l, const void *r ) {
    return eif->comparepp(&l, &r);
}

#endif /* ARR_ELEMENT_H */
