#ifndef ARR_LIST_H
#define ARR_LIST_H

#include "pocofwd.h"
#include "element.h"
#include <stdio.h>
#include <assert.h>

/*
 * User interface for ListNode.
 *
 * Design decision: Does ListNode store a pointer to ElementInterface
 * or not?  I'll assume here that it does thus making each ListNode
 * object a little larger.  This means that the functions don't have
 * to take ElementInterface pointers as arguments.  It might be
 * better to do it the other way.
 */

/*
 * ListNode constructor makes a copy of value and stores a pointer to
 * it.  It also stores a pointer to the ElementInterface, eif, but
 * without making a copy.  Thus, each ListNode "owns" its value but
 * not its ElementInterface.  When a ListNode is destroyed, the value
 * is destroyed too but the ElementInterface is not.
 */
ListNode * ListNode_construct(ElementInterface * eif, void * value);

/*
 * Destroy all memory that is allocated by a single call of
 * ListNode_construct.
 */
void       ListNode_destroy(ListNode * node);

/*
 * Given a ListNode, return a pointer to the next ListNode in the
 * linked list.  Return NULL if node is last in the list.
 */
ListNode * ListNode_next(ListNode *node);

/*
 * Given a ListNode, return a pointer to the previous ListNode in the
 * linked list.  Return NULL if node is first in the list.
 */
ListNode * ListNode_prev(ListNode *node);

/*
 * Return a pointer to the value that is stored at this node.
 */
void     * ListNode_value(ListNode *node);

/*
 * Insert a new ListNode into the linked list immediately AFTER the
 * given node.  Return node.
 */
ListNode * ListNode_insertAfter( ListNode *node, value);

/*
 * Insert a new ListNode into the linked list immediately BEFORE the
 * given node.  Return node.
 */
ListNode * ListNode_insertBefore( ListNode *node, value);

/* 
 * Take a node out of the list and return a pointer to it.  To remove
 * a node altogether without breaking the linked list, call:
 *
 *   ListNode_destroy(ListNode_unlink(node));
 *   node = NULL;
 *
 */
ListNode * ListNode_unlink(ListNode *node);

/*
 * List is built on top of ListNode.  Here is its user interface:
 */

/*
 * Construct a new List.  As with ListInterface, the ElementInterface
 * is not locally owned.
 */
List *     List_construct(ElementInterface * eif);

/*
 * Destroy list along with all its ListNodes.  Does not destroy
 * ElementInterface. 
 */
void       List_destroy(List *l);

/*
 * Insert a ListNode containing given value into a sorted list.  The
 * new ListNode is inserted at the proper position so that the list
 * remains sorted.  If the value is already present in the linked
 * list, then a pointer to the ListNode containing this value is
 * returned.  Otherwise, a new ListNode is constructed and inserted
 * into the list.  Returns NULL on failure.
 */
ListNode * List_insertSorted( List l, void * value);

/*
 * Find the ListNode containing a given value in a sortet List and
 * return a pointer to this ListNode.  Return NULL if value isn't
 * present. 
 */
ListNode * List_findSorted( List l, void * value );

/*
 * Return the ListNode that begins the List, or NULL if the List is
 * empty.
 */
ListNode * List_begin(List * l);

/*
 * Return the ListNode that ends the List, or NULL if the List is
 * empty.
 */
ListNode * List_end(List * l);

/*
 * Add a new ListElement to the end of the list and return a pointer
 * to it.  Return NULL on failure.
 */
ListElement * List_pushBack( List * l, void * value );

/*
 * Add a new ListElement to the front of the list and return a pointer
 * to it.  Return NULL on failure.
 */
ListElement * List_pushFront(void * value, List * l);

#endif /* ARR_LIST_H */
