#ifndef ARR_VECTOR_H
#define ARR_VECTOR_H

#include "poco/pocofwd.h"
#include "poco/element.h"
#include <stdio.h>
#include <assert.h>

/*
 * Vector structure is defined here rather than in vector.c in order
 * to make it possible to define inline functions.
 */

struct Vector {
    /* Pointer to base of vector */
    void ** base;

    /* Number of elements initialized */
    int size;

    /* Allocated length of vector "base". */
    int capacity;

    /* Interface to elements. */
    ElementInterface *eif;
};

inline static void     Vector_append(Vector * target, const Vector * source);
inline static void     Vector_assign( Vector * lhs, const Vector * rhs );
inline static void *   Vector_at( const Vector * v, int i );
inline static int      Vector_capacity(const Vector * v);
inline static void     Vector_clear(Vector * v);
inline static int      Vector_compare(const Vector *l, const Vector *r);
inline static Vector * Vector_construct(ElementInterface *eif);
inline static Vector * Vector_copy(const Vector *old);
inline static void     Vector_destroy (Vector * v);
inline static void *   Vector_export( const Vector * v, int i );
inline static int      Vector_isSorted(Vector *v);
inline static int      Vector_lookup(Vector *v, const void * val);
inline static void    *Vector_popBack( Vector *v);
inline static int      Vector_pushBack( Vector *v, const void *value );
inline static int      Vector_pushBack_noCopy( Vector *v, void *value );
int                    Vector_reserve( Vector * v, int n );
int                    Vector_resize( Vector * v, int n );
inline static void     Vector_setValue( Vector * v, int i, void * newValue );
inline static void     Vector_setPtr_noCopy( Vector * v, int i,
					     void * newPtr );
inline static int      Vector_size(const Vector * v);
inline static void     Vector_sort(Vector * v);

/** Append a copy of source to target */
inline static void     
Vector_append(Vector * target, const Vector * source) {
    int i;

    for(i=0; i < Vector_size(source); ++i)
	Vector_pushBack( target, Vector_at(source, i) );
}

/*
 * Make lhs a copy of rhs.
 */
inline static void
Vector_assign( Vector * lhs, const Vector * rhs ) {
    int i;
    if(lhs != rhs) {
	assert(lhs);
	assert(rhs);

	/*
	 * I assume without checking that lhs->eif
	 * and rhs->eif are identical.
	 */

	/* Should use whichever of the following is faster */
#if 0
	/* probably slower */
	Vector_clear(lhs);
	Vector_reserve(lhs, rhs->capacity);
	for(i=0; i < rhs->size; ++i)
	    Vector_pushBack(lhs, Vector_at(rhs, i) );
#else
	/* probably faster */
	Vector_resize(lhs, rhs->size);
	for(i=0; i < rhs->size; ++i) {
	    lhs->base[i] = lhs->eif->assign(lhs->base[i], rhs->base[i]);
	}
#endif
    }
}

/*
 * Return a pointer to a copy of vector "old".
 */
inline static Vector * 
Vector_copy(const Vector *old) {
    Vector *new = Vector_construct(old->eif);
    Vector_assign(new, old);
    return new;
}


/* Return pointer to i'th value */
inline static void *
Vector_at( const Vector * v, int i ){
    assert(v != NULL);
    assert(i < v->size);
    return v->base[i];
}

/** How many values can the vector hold without reallocating? */
inline static int      
Vector_capacity(const Vector * v) {
    assert(v != NULL);
    return(v->capacity);
}

/*
 * Return 1, 0, or -1 to indicate that l is less than, equal to, or
 * greater than r.  Mnemonically, the function returns (l-r).
 */
inline static int      
Vector_compare(const Vector * l, const Vector *r) {
    if(l != r) {
	int i, c, minsize;

	assert(l);
	assert(r);

	minsize = (l->size < r->size) ? l->size : r->size;
    
	/* is there a difference in element values? */
	for(i = 0; i < minsize; ++i) {
	    c = ElementInterface_compare(l->eif, l->base[i], r->base[i]);
	    if(c)
		return c;
	}

	/* No difference in values.  Is there a difference in length? */
	if(l->size < r->size)  return 1;
	if(l->size > r->size)  return -1;
    }
    return 0;
}

inline static Vector *
Vector_construct (ElementInterface *eif) {
    Vector * v = (Vector * ) malloc(sizeof(Vector));
    if(v == NULL)
	return NULL;
    v->base = NULL;
    v->size = v->capacity = 0;
    v->eif = ElementInterface_copy(eif);

    return v;
}

/* Set size to zero, destroying any existing values */
inline static void
Vector_clear(Vector * v) {
    Vector_resize( v, 0 );
}

/** Return i'th ptr and set that ptr to NULL **/
inline static void *   
Vector_export( const Vector * v, int i ) {
    void * rval;
    assert(v != NULL);
    assert(i < v->size);
    rval = v->base[i];
    v->base[i] = NULL;
    return rval;
}

inline static int
Vector_isSorted(Vector *v) {
    int i;

    for(i=0; i < v->size - 1; ++i) {
	if(ElementInterface_compare(v->eif, v->base[i], v->base[i+1]) > 0) {
	    return 0;
	}
    }
    return 1;
}

/**
 * Vector v must be sorted in ascending order before this function is
 * called.  Function returns index of smallest element in sorted array
 * that is >= val.  Before the algorithm begins, the function checks
 * that vec[lo] <= vec[hi] and that val <= vec[hi].  If this condition
 * fails, then the function returns -1.  Otherwise, it assumes that v
 * is sorted and proceeds.
 */
inline static int
Vector_lookup(Vector *v, const void * val) {
    int             lo, mid, hi;

    lo = 0;
    hi = v->size - 1;
    if(!(ElementInterface_compare(v->eif, v->base[lo], v->base[hi]) <= 0
	 && ElementInterface_compare(v->eif, val, v->base[hi]) <= 0))
        return -1;
    if(hi == 0)
        return 0;
    while(lo < hi) {
        mid = lo + (hi - lo) / 2;
        if(mid == lo)
            break;
        if(ElementInterface_compare(v->eif, v->base[mid], val) < 0)
            lo = mid;
        else
            hi = mid;
    }
    if(ElementInterface_compare(v->eif, v->base[lo], val) >=0)
        hi = lo;

    assert(hi >= 0);
    assert(hi < v->size);
    assert(ElementInterface_compare(v->eif, v->base[hi], val) >= 0);
    assert(hi == 0
	   || ElementInterface_compare(v->eif,v->base[hi - 1], val) < 0);

    return hi;
}

/*
 * Push a (copy of a) new value onto the end of the vector.
 * Increases size by 1.
 * Increases capacity if size==capacity.
 */
inline static int
Vector_pushBack( Vector *v, const void *value ) {
    assert(v != NULL);

    return Vector_pushBack_noCopy(v, v->eif->copy(value));
}

/*
 * Push a new value onto the end of the vector.
 * Increases size by 1.
 * Increases capacity if size==capacity.
 */
inline static int
Vector_pushBack_noCopy( Vector *v, void *value ) {
    assert(v != NULL);

    if(v->size == v->capacity) {
	if( Vector_reserve( v, v->size + 1 ) != 0)
	    return 1;
    }
    v->base[v->size] = value;
    ++(v->size);
    return 0;
}

/*
 * Return a pointer to the last element in the vector, then reduce
 * the vector's size by 1.  Return NULL if the vector is empty.
 */
inline static void *
Vector_popBack( Vector *v) {
    void *ptr;

    if(v->size == 0)
	return NULL;

    ptr = v->base[v->size-1];
    --v->size;
    return ptr;
}

inline static void
Vector_destroy (Vector * v) {
    int i;
    assert(v != NULL);

    for(i=0; i < v->size; ++i)
	v->eif->destroy(v->base[i]);
    if(v->base)
	free(v->base);
    ElementInterface_destroy(v->eif);
    free(v);
}

/*
 * How many values are in the vector?
 */
inline static int
Vector_size(const Vector * v) {
    assert(v != NULL);
    return(v->size);
}

/*
 * Assign newValue to the i'th entry of vector v.
 */
inline static void
Vector_setValue( Vector * v, int i, void * newValue ) {
    assert(v != NULL);
    assert(i >= 0);
    assert(i < v->size);
    if(v->base[i])
	v->base[i] = v->eif->assign(v->base[i], newValue);
    else
	v->base[i] = v->eif->copy(newValue);
}

/*
 * Assign new value to the i'th pointer without copying the object to
 * which the pointer points.  The target pointer must previously have
 * been set to NULL;
 */
inline static void 
Vector_setPtr_noCopy( Vector * v, int i, void * newPtr ) {
    assert(v != NULL);
    assert(i < v->size);
    assert(v->base[i] == NULL);
    v->base[i] = newPtr;
}

inline static void
Vector_sort(Vector * v) {
    qsort(v->base, (unsigned) v->size, sizeof(v->base[0]), v->eif->comparepp);
}

#endif /* ARR_VECTOR_H */
