#include "poco/bintree.h"
#include <stdlib.h>

/* static functions define the private interface of BinTreeNode */
static BinTreeNode *BinTreeNode_insert (BinTreeNode * currNode,
					BinTreeNode * parent, 
					BinTreeNode **insertedAt,
					ElementInterface *eif,
					const void *key );
static BinTreeNode *BinTreeNode_find (BinTreeNode * node,
				      ElementInterface *eif,
				      const void *key );

struct BinTreeNode {
    void *value;
    int count;
    struct BinTreeNode *left, *right, *up;
};

struct BinTree {
    /* Pointer to root of binary tree */
    BinTreeNode *root;

    /*
     * Pointer to structure defining interface to elements.
     */
    ElementInterface *eif;
};

/**
 * Insert a new key in a binary tree of BinTreeNodes.  If the key
 * already exists, then increment its count.  If node==NULL on input,
 * then return a pointer to a newly allocated node.  Otherwise return
 * node.  Function returns NULL on failure.
 *
 * @param currNode The current node of the binary tree, or NULL if *key is
 * not yet present in the tree.
 *
 * @param parent The node that is ancestral to currNode in the sense
 * that it is closer to the root of the tree.
 *
 * @param insertedAt The address of a BinTreeNode pointer in the calling
 * function.  On return, this will point to the BinTreeNode at which
 * the key was inserted.
 *
 * @param eif The ElementInterface that is used to manipulate the key.
 *
 * @param key The value that is to be inserted in the tree.
 **/
static BinTreeNode *
BinTreeNode_insert (BinTreeNode * currNode, BinTreeNode * parent,
		    BinTreeNode **insertedAt,
		    ElementInterface *eif, const void *key )
{
  int diff;

  if (currNode == NULL) {	     /* allocate a BinTreeNode */
    currNode = (BinTreeNode *) malloc (sizeof (BinTreeNode));
    if(currNode == NULL)
      return NULL;
    currNode->value = eif->copy(key);
    currNode->count = 1;
    currNode->up = parent;
    currNode->left = currNode->right = NULL;
    *insertedAt = currNode;
    return currNode;
  }

  diff = ElementInterface_compare(eif, key, currNode->value);
  if (diff < 0) {
    currNode->left = BinTreeNode_insert (currNode->left, currNode,
					 insertedAt, eif, key); 
    if (currNode->left == NULL)
      return NULL;
  } else if (diff > 0) {
    currNode->right = BinTreeNode_insert (currNode->right, currNode,
					  insertedAt, eif, key);
    if (currNode->right == NULL)
      return NULL;
  } else {         /* diff == 0 */ 
    ++currNode->count;
    *insertedAt = currNode;
  }

  return currNode;
}

/** Return a pointer to the BinTreeNode whose value matches key, or NULL if
    no such node exists. **/
static BinTreeNode *
BinTreeNode_find ( BinTreeNode * node, ElementInterface *eif, const void *key )
{
    int diff;

    if (node == NULL)
	return NULL;

    diff = ElementInterface_compare(eif, key, node->value);
    if (diff < 0)
	return BinTreeNode_find (node->left, eif, key);
    else if (diff > 0)
	return BinTreeNode_find (node->right, eif, key);
    /* else diff == 0 */
    return node;
}

void
BinTreeNode_destroy ( BinTreeNode * bt, ElementInterface * eif )
{
    if (bt == NULL)
	return;

    /* delete children */
    BinTreeNode_destroy ( bt->left, eif );
    BinTreeNode_destroy ( bt->right, eif );

    /* delete self */
    eif->destroy(bt->value);
    free (bt);
}

/* Return value of count variable */
int
BinTreeNode_count (BinTreeNode * node)
{
    return node->count;
}

/* Return pointer to data variable */
void *
BinTreeNode_value (BinTreeNode * node)
{
    return node->value;
}

/** Return pointer to minimum BinTreeNode in binary tree */
BinTreeNode *
BinTreeNode_min (BinTreeNode * node)
{
    if (node == NULL)
	return NULL;
    while (node->left != NULL)
	node = node->left;
    return node;
}

/** Return pointer to next BinTreeNode in sequence */
BinTreeNode *
BinTreeNode_next (BinTreeNode * node)
{
    BinTreeNode *last;

    if (node == NULL)
	return NULL;
    if (node->right != NULL)
	return BinTreeNode_min (node->right);
    else
	do {
	    /* go up while we are parent's right child */
	    if (node->up == NULL)
		return NULL;
	    last = node;
	    node = node->up;
	} while (node->right == last);
    return node;
}

/**
  Allocate an empty BinTree object.  Return pointer to allocated
  BinTree on success, or NULL on failure.

  In ordinary usage, data values are copied as they are inserted into
  the BinTree and these copies are deleted when the BinTree is
  deleted.  It is also possible, however, for data values to be owned
  by the calling program and be neither copied nor deleted.  For this
  case, define the copy and delete functions within eif as follows:

    void * copy(void * rhs) { return rhs; }
    void delete(void * obj) {}

  @param eif The ElementInterface, an object that defines the
  functions that are used for manipulating the data objects that are
  stored in the tree.
**/
BinTree *
BinTree_construct (ElementInterface *eif)
{
    BinTree *new = (BinTree *) malloc (sizeof (BinTree));
    if(new == NULL)
      return NULL;
    new->eif = ElementInterface_copy(eif);
    new->root = NULL;
    return new;
}

/**
 * Insert value "key" in binary tree, incrementing count if key is already
 * there.  On success, return a const pointer to the BinTreeNode containing
 * the key.  On failure, return NULL.
 *
 * The value returned by this function is const, and should not be
 * handed to free.  It will be freed by BinTree_destroy.
 */
const BinTreeNode *
BinTree_insert ( BinTree * bt, const void *key )
{
  BinTreeNode *insertedAt;
  if (bt == NULL)
    return 1;                               /* failure */
  
  bt->root = BinTreeNode_insert( bt->root, NULL, &insertedAt, bt->eif, key );
  
  if (bt->root == NULL)
    return NULL;				      /* failure */
  return insertedAt;
}

/** Return a pointer to the BinTreeNode whose value matches key, or NULL if
    no such node exists. **/
BinTreeNode *
BinTree_find ( const BinTree * bt, const void *key )
{
    return BinTreeNode_find ( bt->root, bt->eif, key );
}

void
BinTree_destroy (BinTree * bt)
{
    if (bt == NULL)
	return;

    BinTreeNode_destroy ( bt->root, bt->eif );
    ElementInterface_destroy(bt->eif);
    free(bt);
}

/** Find minimum node */
BinTreeNode * 
BinTree_min(const BinTree * bt)
{
    return BinTreeNode_min(bt->root);
}
