#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "poco/dblvec.h"

struct DblVec {
    double * base;       /* Pointer to base of vector */
    int size;       /* Number of elements initialized */
    int capacity;   /* Allocated length of vector "base". */
};

static int dblcompar(const double *i, const double *j);

/** comparison function for qsort **/
static int 
dblcompar(const double *i, const double *j) {
    if(*i < *j)
	return -1;
    if(*i == *j)
	return 0;
    return 1;
}

void
DblVec_append(DblVec * target, const DblVec * source) {
    int i;

    for(i=0; i < DblVec_size(source); ++i)
	DblVec_pushBack( target, DblVec_at(source, i) );
}

void
DblVec_assign(DblVec * l, const DblVec * r) {
    int i;
    if(l != r) {
	assert(l);
	assert(r);
	DblVec_clear(l);
	DblVec_reserve(l, r->capacity);
	for(i=0; i < r->size; ++i)
	    DblVec_pushBack(l, DblVec_at(r, i));
    }
}

/*
 * Assign newValue to the i'th entry of vector v.
 */
void
DblVec_setValue( DblVec * v, int i, double newValue) {
    assert(v != NULL);
    assert(i < v->size);
    v->base[i] = newValue;
}

double
DblVec_at( const DblVec * v, int i){
    assert(v != NULL);
    assert(i < v->size);
    return v->base[i];
}

/*
 * How many values can the vector hold without reallocating?
 */
int
DblVec_capacity(const DblVec * v) {
    assert(v != NULL);
    return(v->capacity);
}

/* Set size to zero, destroying any existing values */
void
DblVec_clear(DblVec * v) {
    DblVec_resize( v, 0);
}

/*
 * lp and rp point to the ADDRESSES OF data values.
 */
int      
DblVec_comparepp(DblVec * const *lpp, DblVec * const *rpp) {
    return DblVec_compare(*lpp, *rpp);
}

/*
 * lp and rp point to data values.
 */
int      
DblVec_compare(const DblVec *l, const DblVec *r) {
    int i, minsize;

    if(l == r)
	return 0;

    assert(l);
    assert(r);

    minsize = (l->size < r->size ? l->size : r->size);
    for(i=0; i < minsize; ++i) {
	if(l->base[i] > r->base[i]) {
	    return 1;
	}else if(l->base[i] < r->base[i])
	    return -1;
    }
    if(l->size > r->size)
	return 1;

    if(l->size < r->size)
	return -1;

    return 0;
}

DblVec *
DblVec_construct (int size) {
    DblVec * v = (DblVec * ) malloc(sizeof(*v));
    if(v == NULL)
	return NULL;
    v->base = NULL;
    v->size = v->capacity = 0;
    DblVec_reserve( v, size);
    return v;
}

DblVec *
DblVec_copy(const DblVec * v) {
    DblVec * new = DblVec_construct(v->size);
    if(new==NULL)
	return NULL;
    memcpy(new->base, v->base, v->size * sizeof( new->base[0] ));
    new->size = v->size;
    return new;
}

void
DblVec_destroy (DblVec * v) {
    assert(v != NULL);
    if(v->base)
	free(v->base);
    free(v);
}

void
DblVec_increment( DblVec *v, int ndx ) {
    assert(v);
    assert(ndx < v->size);
    ++(v->base[ndx]);
}

/*
 * Return nonzero if vector is sorted in ascending order; zero otherwise.
 */
int      
DblVec_isSorted(DblVec *v) {
    int i;

    for(i=0; i < v->size - 1; ++i) {
	if(v->base[i] > v->base[i+1]) {
	    return 0;
	}
    }
    return 1;
}

/**
 * Vector v must be sorted in ascending order before this function is
 * called.  Function returns index of smallest element in sorted array
 * that is >= val.  Before the algorithm begins, the function checks
 * that vec[lo] <= vec[hi] and that val <= vec[hi].  If this condition
 * fails, then the function returns -1.  Otherwise, it assumes that v
 * is sorted and proceeds.
 */
int
DblVec_lookup(DblVec *v, double val) {
    int             lo, mid, hi;

    lo = 0;
    hi = v->size - 1;
    if(!(v->base[hi] >= v->base[lo] && val <= v->base[hi]))
        return -1;
    if(hi == 0)
        return 0;
    while(lo < hi) {
        mid = lo + (hi - lo) / 2;
        if(mid == lo)
            break;
        if(v->base[mid] < val)
            lo = mid;
        else
            hi = mid;
    }
    if(v->base[lo] >= val)
        hi = lo;

    assert(hi >= 0);
    assert(hi < v->size);
    assert(v->base[hi] >= val);
    assert(hi == 0 || v->base[hi - 1] < val);

    return hi;
}

/*
 * Return a pointer to the last element in the vector, then reduce
 * the vector's size by 1.  Abort if the vector is empty.
 */
double   DblVec_popBack(DblVec *v) {
    double rval;

    if(v->size == 0) {
	fprintf(stderr, "Can't pop an empty DblVec @%s:%d\n",
		__FILE__,__LINE__);
	exit(1);
    }

    rval = v->base[v->size-1];
    --v->size;
    return rval;
}

void     
DblVec_print(const DblVec *v, FILE *fp) {
    int i;

    assert(fp);
    assert(v);

    for(i=0; i < v->size; ++i)
	fprintf(fp," [%d, %g]", i, v->base[i]);
    putc('\n', fp);
}

/*
 * Push a (copy of a) new value onto the end of the vector.
 * Increases size by 1.
 * Increases capacity if size==capacity.
 */
int
DblVec_pushBack( DblVec *v, double value ) {
    assert(v != NULL);

    if(v->size == v->capacity) {
	if( DblVec_reserve( v, v->size + 1 ) != 0)
	    return 1;
    }
    v->base[v->size] = value;
    ++(v->size);
    return 0;
}

/*
 * Resize vector to new size n, copying contents to new location if
 * necessary.  Return 0 on success; nonzero on failure
 */
int
DblVec_resize( DblVec * v, int n ) {
    int i;

    if(n <= v->size) {
	/* no-op if n == size */
	v->size = n;
	return 0;
    }

    /* else n > v->size */
    if(n > v->capacity) {
	if( DblVec_reserve( v, n ) != 0 )
	    return 1;
    }
    for(i=v->size; i < n; ++i)
	v->base[i] = 0;
    v->size = n;
    return 0;
}

/*
 * Increase capacity, the number of values allocated.
 * No-op if n < current capacity.
 * Return zero on success, non-zero on failure.
 */
int
DblVec_reserve( DblVec * v, int n ) {
    if(n <= v->capacity)
	return 0;

    v->base = (double *) realloc(v->base, n * sizeof(v->base[0]));
    if(v->base == NULL) {
	v->size = v->capacity = 0;
	return 1;
    }
    v->capacity = n;
    return 0;
}

/** How many values are in the vector? **/
int
DblVec_size(const DblVec * v) {
    assert(v != NULL);
    return v->size;
}

void
DblVec_sort(DblVec * v) {
    qsort(v->base, (unsigned) v->size, sizeof(v->base[0]),
	  (int(*)(const void *, const void *)) dblcompar);
}

/** Return the sum of the elements */
double   
DblVec_sum(const DblVec *v) {
    int i;
    double sum=0.0;

    for(i=0; i < v->size; ++i)
	sum += v->base[i];

    return sum;
}
