#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "poco/intvec.h"

struct IntVec {
    int * base;       /* Pointer to base of vector */
    int size;       /* Number of elements initialized */
    int capacity;   /* Allocated length of vector "base". */
};

static int icompar(const int *i, const int *j);

/** comparison function for qsort **/
static int 
icompar(const int *i, const int *j) {
    return *i - *j;
}

/** Append a copy of source to target */
void
IntVec_append(IntVec * target, const IntVec * source) {
    int i;

    for(i=0; i < IntVec_size(source); ++i)
	IntVec_pushBack( target, IntVec_at(source, i) );
}

void     
IntVec_assign(IntVec * lhs, const IntVec * rhs ) {
    int i;
    if(lhs != rhs) {
	assert(lhs);
	assert(rhs);
	IntVec_clear(lhs);
	IntVec_reserve(lhs, rhs->capacity);
	for(i=0; i < rhs->size; ++i)
	    IntVec_pushBack(lhs, IntVec_at(rhs, i) );
    }
}

/*
 * Assign newValue to the i'th entry of vector v.
 */
void
IntVec_setValue(IntVec * v, int i, int newValue) {
    assert(v != NULL);
    assert(i < v->size);
    v->base[i] = newValue;
}

int
IntVec_at(const IntVec * v, int i){
    assert(v != NULL);
    assert(i < v->size);
    return v->base[i];
}

/** How many values can the vector hold without reallocating? */
int
IntVec_capacity(const IntVec * v) {
    assert(v != NULL);
    return(v->capacity);
}

/* Set size to zero, destroying any existing values */
void
IntVec_clear(IntVec * v) {
    IntVec_resize(v, 0);
}

IntVec *
IntVec_construct (int capacity) {
    IntVec * v = (IntVec * ) malloc(sizeof(*v));
    if(v == NULL)
	return NULL;
    v->base = NULL;
    v->size = v->capacity = 0;
    IntVec_reserve(v, capacity);
    return v;
}

IntVec *
IntVec_copy(const IntVec * v) {
    IntVec * new = IntVec_construct(v->size);
    if(new==NULL)
	return NULL;
    memcpy(new->base, v->base, v->size * sizeof( new->base[0] ));
    new->size = v->size;
    return new;
}

void
IntVec_destroy (IntVec * v) {
    assert(v);
    if(v->base) {
	assert(v->capacity > 0);
	free(v->base);
    }
    free(v);
}

void
IntVec_increment(IntVec *v, int ndx, int inc) {
    assert(v);
    if(ndx >= v->size)
	IntVec_resize(v, ndx+1);
    v->base[ndx] += inc;
}

/*
 * Return nonzero if vector is sorted in ascending order; zero otherwise.
 */
int      
IntVec_isSorted(IntVec *v) {
    int i;

    for(i=0; i < v->size - 1; ++i) {
	if(v->base[i] > v->base[i+1]) {
	    return 0;
	}
    }
    return 1;
}

/**
 * Vector v must be sorted in ascending order before this function is
 * called.  Function returns index of smallest element in sorted array
 * that is >= val.  Before the algorithm begins, the function checks
 * that vec[lo] <= vec[hi] and that val <= vec[hi].  If this condition
 * fails, then the function returns -1.  Otherwise, it assumes that v
 * is sorted and proceeds.
 */
int      
IntVec_lookup(IntVec *v, int val) {
    int             lo, mid, hi;

    lo = 0;
    hi = v->size - 1;
    if(!(v->base[lo] <= v->base[hi] && val <= v->base[hi]))
        return -1;
    if(hi == 0)
        return 0;
    while(lo < hi) {
        mid = lo + (hi - lo) / 2;
        if(mid == lo)
            break;
        if(v->base[mid] < val)
            lo = mid;
        else
            hi = mid;
    }
    if(v->base[lo] >= val)
        hi = lo;

    assert(hi >= 0);
    assert(hi < v->size);
    assert(v->base[hi] >= val);
    assert(hi == 0 || v->base[hi - 1] < val);

    return hi;
}

/*
 * Return a pointer to the last element in the vector, then reduce
 * the vector's size by 1.  Abort if the vector is empty.
 */
int
IntVec_popBack(IntVec *v) {
    int rval;

    if(v->size == 0) {
	fprintf(stderr, "Can't pop an empty IntVec @%s:%d\n",
		__FILE__,__LINE__);
	exit(1);
    }

    rval = v->base[v->size-1];
    --v->size;
    return rval;
}

/*
 * Push a (copy of a) new value onto the end of the vector.
 * Increases size by 1.
 * Increases capacity if size==capacity.
 */
int
IntVec_pushBack(IntVec *v, int value) {
    assert(v != NULL);

    if(v->size == v->capacity) {
	if( IntVec_reserve(v, v->size + 1) != 0)
	    return 1;
    }
    v->base[v->size] = value;
    ++(v->size);
    return 0;
}

/*
 * Resize vector to new size n, copying contents to new location if
 * necessary.  Return 0 on success; nonzero on failure
 */
int
IntVec_resize(IntVec * v, int n) {
    if(n <= v->size) {
	/* no-op if n == size */
	v->size = n;
	return 0;
    }

    /* else n > v->size */
    if(n > v->capacity) {
	if( IntVec_reserve(v, n) != 0 )
	    return 1;
    }
    memset(v->base + v->size, 0, (n - v->size)*sizeof(v->base[0]));
    v->size = n;
    return 0;
}

/*
 * Increase capacity, the number of values allocated.
 * No-op if n < current capacity.
 * Return zero on success; abort on failure.
 */
int
IntVec_reserve(IntVec * v, int n) {
    if(n <= v->capacity)
	return 0;

    v->base = (int *) realloc(v->base, n * sizeof(v->base[0]));
    if(v->base == NULL) {
	fprintf(stderr,"Failed attempt to realloc %d bytes @%s:%d",
		n * sizeof(v->base[0]),
		__FILE__, __LINE__);
	exit(1);
    }
    v->capacity = n;
    return 0;
}

/*
 * How many values are in the vector?
 */
int
IntVec_size(const IntVec * v) {
    assert(v);
    return v->size;
}

void     
IntVec_print(const IntVec *v, FILE *fp) {
    int i;

    assert(fp);
    assert(v);

    for(i=0; i < v->size; ++i)
	fprintf(fp," [%d, %d]", i, v->base[i]);
    putc('\n', fp);
}

/*
 * Compare two IntVec structures.  Returns -1, 0, or 1 depending on
 * whether l is less than, equal to, or greater than r.
 *
 * lp and rp point to the ADDRESSES OF data values.
 */
int
IntVec_comparepp(IntVec * const *lp, IntVec * const *rp) {
    return IntVec_compare(*lp, *rp);
}

/*
 * Compare two IntVec structures.  Returns -1, 0, or 1 depending on
 * whether l is less than, equal to, or greater than r.
 *
 * l and r point to data values.
 */
int
IntVec_compare(const IntVec *l, const IntVec *r) {
    if(l != r) {
	int i, minsize;

	minsize = (l->size < r->size) ? l->size : r->size;
    
	/* is there a difference in element values? */
	for(i = 0; i < minsize; ++i) {
	    if(l->base[i] < r->base[i])  return 1;
	    if(l->base[i] > r->base[i])  return -1;
	}

	/* No difference in values.  Is there a difference in length? */
	if(l->size < r->size)  return 1;
	if(l->size > r->size)  return -1;
    }
    return 0;
}

void
IntVec_sort(IntVec * v) {
    qsort(v->base, (unsigned) v->size, sizeof(v->base[0]),
	  (int(*)(const void *, const void *)) icompar);
}

/** Return the sum of the elements */
int
IntVec_sum(const IntVec *v) {
    int i, sum=0;

    for(i=0; i < v->size; ++i)
	sum += v->base[i];

    return sum;
}
