/**
   @file ndxlbl.c

   @brief Functions manipulating structure NdxLbl, which keeps track of the
   relationship between labels and the indices that refer to them. 
   
   @internal
   Copyright (C) 2003 Alan R. Rogers

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Alan R. Rogers, Department of Anthropology, University of Utah,
   Salt Lake City, UT 84112. rogers@anthro.utah.edu
**/
#include "poco/ndxlbl.h"
#include <string.h>
#include <stdlib.h>
#include <assert.h>

typedef struct {
    char * lbl;
    int ndx;
} NdxLblNode;

struct NdxLbl {
    BinTree * bt;      /* binary tree */
    int size;
};

static NdxLblNode * NdxLblNode_construct(const char * lbl, int ndx);
static void NdxLblNode_destroy(void * node);
static void * NdxLblNode_copy(const void *rhs);
static int NdxLblNode_comparepp(const void * lhs, const void * rhs);

static ElementInterface * ndxLblNodeInterface = NULL;

static NdxLblNode *
NdxLblNode_construct(const char * lbl, int ndx)
{
    NdxLblNode * new = (NdxLblNode *) malloc(sizeof(NdxLblNode));
    if(new == NULL)
	return NULL;
    new->lbl = strdup(lbl);
    new->ndx = ndx;
    return new;
}

static void
NdxLblNode_destroy(void * node_void)
{
    NdxLblNode * node = (NdxLblNode *) node_void;
    assert(node_void != NULL);
    free(node->lbl);
    free(node);
}

static void *
NdxLblNode_copy(const void *rhs_void)
{
    const NdxLblNode * rhs = (const NdxLblNode *) rhs_void;
    NdxLblNode * lhs = (NdxLblNode *) malloc(sizeof(NdxLblNode));
    if(lhs == NULL)
	return NULL;
    
    lhs->lbl = strdup(rhs->lbl);
    lhs->ndx = rhs->ndx;
    return lhs;
}

static int
NdxLblNode_comparepp(const void *lpp, const void *rpp)
{
    const NdxLblNode * l = *((NdxLblNode * const *) lpp);
    const NdxLblNode * r = *((NdxLblNode * const *) rpp);
    /* does not compare ndx variables of lhs and rhs */
    return strcmp(l->lbl, r->lbl);
}

/** default constructor */
NdxLbl *
NdxLbl_construct(void)
{
    NdxLbl * nl;

    /* ndxLblNodeInterface is constructed on 1st call */
    if(ndxLblNodeInterface == NULL) {
	ndxLblNodeInterface
	    = ElementInterface_construct(NULL,
					 NdxLblNode_comparepp,
					 NULL,
					 NdxLblNode_copy,
					 NdxLblNode_destroy,
					 NULL
					 );
	if(ndxLblNodeInterface == NULL)
	    return NULL;
    }

    nl = (NdxLbl *) malloc(sizeof(NdxLbl));
    if(nl == NULL)
	return NULL;

    nl->bt = BinTree_construct(ndxLblNodeInterface);
    if(nl->bt == NULL) {
	free(nl);
	return NULL;
    }
    nl->size = 0;

    return nl;
}

void
NdxLbl_destroy(NdxLbl * nl)
{
    assert(nl != NULL);
    BinTree_destroy(nl->bt);
    free(nl);
}

/** 
 * Insert lbl into NdxLbl object, incrementing count if lbl is already
 * there.  
 * @return the index of lbl (a non-negative integer) on success, or -1
 * on failure.
 */
int
NdxLbl_insert(NdxLbl * nl, const char * lbl)
{
    NdxLblNode * nln;
    BinTreeNode * btn;

    assert(nl != NULL);
    
    /* Create a new NdxLblNode, with ndx variable equal to
     * the current size of nl.  The ndx variable won't be
     * used if lbl is already present in nl. */
    nln = NdxLblNode_construct(lbl, nl->size);

    /* Find the BinTreeNode whose value equals nln.  In this comparison,
     * only the lbl member variables are considered.
     */
    btn = BinTree_find( nl->bt, nln );
    if(btn == NULL) 
	++nl->size;

    /*
     * Add nln to nl.   NdxLblNode objects compare as equal provided only
     * that their lbl member variables are equal.  The ndx member variables
     * are ignored.  Consequently, nln will be added as a new node in nl
     * only if label lbl is not already present there.  In that case, the
     * ndx member variable of the new node is nl->size as specified just
     * above.  If lbl is already present in nl, its associated ndx variable
     * is not altered.
     */
    btn = BinTree_insert( nl->bt, nln );
    if(btn == NULL)
      return -1;      /* failure */

    nln = BinTreeNode_value(btn);

    return nln->ndx;  /* success */
}

/**
 * Find index of label lbl.
 * @param lbl A pointer to a string containing a label.
 * @return index of lbl (if found) or -1 (if lbl isn't there).
 **/
int
NdxLbl_ndx(NdxLbl *nl, const char * lbl)
{
    NdxLblNode * nln;
    BinTreeNode * btn;

    assert(nl != NULL);
    
    /* Create a new NdxLblNode. 2nd argument is arbitrary */
    nln = NdxLblNode_construct(lbl, 0);

    /* Find the BinTreeNode whose value equals nln.  In this comparison,
     * only the lbl member variables are considered. */
    btn = BinTree_find( nl->bt, nln );

    NdxLblNode_destroy(nln);
    nln = NULL;

    if(btn == NULL)
	return  -1;

    nln = (NdxLblNode *) BinTreeNode_value(btn);

    /* Return the ndx associated with lbl.*/
    return nln->ndx;
}

/**
 * @param ndx the (integer) index of a group
 * @return a pointer to a newly allocated string containing the group
 * label associated with ndx.  If ndx isn't found, 0 is returned.
 **/
char *
NdxLbl_lbl(NdxLbl *nl, int ndx)
{
    NdxLblNode * nln;
    BinTreeNode * btn;

    assert(nl != NULL);

    /* Find lbl of ndx by walking through tree */
    btn = BinTree_min(nl->bt);
    while(btn != NULL) {
	nln = (NdxLblNode *) BinTreeNode_value(btn);
	if(nln->ndx == ndx)
	    return strdup(nln->lbl);
	btn = BinTreeNode_next(btn);
    }
    return NULL;  /* ndx is not in nl */
}

/** Return the number of observations with given label */
int
NdxLbl_countLbl(NdxLbl *nl, const char * lbl)
{
    NdxLblNode * nln;
    BinTreeNode * btn;

    assert(nl != NULL);
    
    /* Create a new NdxLblNode.  Value of 2nd argument doesn't matter */
    nln = NdxLblNode_construct(lbl, 0);

    /* Find the BinTreeNode whose value equals nln.  In this comparison,
     * only the lbl member variables are considered.  Thus, we end up
     * with the node whose lbl member variable equals the lbl variable
     * that was passed to this function as an argument.  */
    btn = BinTree_find( nl->bt, nln );
    if(btn == NULL) {
	/* lbl not found: count=0 */
	NdxLblNode_destroy(nln);
	return 0;
    }
    
    NdxLblNode_destroy(nln);
    return BinTreeNode_count(btn);
}

/** Return the number of observations with given ndx */
int
NdxLbl_countNdx(NdxLbl *nl, int ndx)
{
    NdxLblNode * nln;
    BinTreeNode * btn;

    assert(nl != NULL);

    /* Find lbl of ndx by walking through tree */
    btn = BinTree_min(nl->bt);
    while(btn != NULL) {
	nln = (NdxLblNode *) BinTreeNode_value(btn);
	if(nln->ndx == ndx)
	    return BinTreeNode_count(btn);
	btn = BinTreeNode_next(btn);
    }
    return 0;  /* ndx is not in nl */
}

/** Return the number of labels stored */
int
NdxLbl_size(NdxLbl *nl)
{
    return nl->size;
}

