#include "poco/vector.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/* Test Vector with data type string */
void * strAssign(void * lhs, const void * rhs);
void * strConstruct(void *dummy);
void * strCpy(const void * data);
int strCompare(const void *lpp, const  void *rpp);
void strDestroy(void * s);
void tstElement(int verbose);

void * strConstruct(void *dummy) {
    char * s = (char *) malloc( sizeof(char) );
    if(s==NULL) {
	fprintf(stderr,"Bad malloc @%s:%d", __FILE__, __LINE__);
	exit(1);
    }
    *s = '\0';
    assert(dummy==NULL);
    return (void *) s;
}

void * strAssign(void * lv, const void * rv) {
    char * lhs = (char *) lv;
    const char * rhs = (const char *) rv;
    size_t len = strlen(rhs);
    if(strlen(lhs) < len) {
	free(lhs);
	lhs = malloc( (len+1)*sizeof(char));
	if(lhs == NULL) {
	    fprintf(stderr,"Bad malloc @%s:%d", __FILE__, __LINE__);
	    exit(1);
	}
    }
    strncpy(lhs, rhs, len+1);
    return lhs;
}

void * strCpy(const void * data)
{
    char * to;
    const char * from = (const char *) data;
    if(from == NULL)
	return NULL;

    to = strdup(from);

    return (void *) to;
}

int strCompare(const void *lpp, const void *rpp)
{
    const char * l = *((char * const *) lpp);
    const char * r = *((char * const *) rpp);
    return strcmp(l,r);
}

void strDestroy(void * s)
{
    free(s);
}

/** Test Vector.  Use -v arg for printed output. **/
int main(int argc, char ** argv)
{
    int verbose = 0;
    int i;
    void *p;
    ElementInterface * eif = ElementInterface_construct(strAssign,
							strCompare,
							strConstruct,
							strCpy,
							strDestroy,
							NULL);
    Vector * v1, * v2, * vtmp1, * vtmp2;
    char * str1 = strdup("test string 1");
    char * str2 = strdup("test string 2");
    char * str3 = strdup("test string 3");

    if(argc > 1 && strcmp(argv[1], "-v") == 0)
	verbose = 1;

    v1 = Vector_construct(eif);
    Vector_destroy(v1);

    v1 = Vector_construct(eif);
    v2 = Vector_construct(eif);

    assert(Vector_size(v1) == 0);
    assert(Vector_capacity(v1) == 0);

    Vector_pushBack( v1, str1 );

    assert(Vector_size(v1) == 1);
    assert(Vector_capacity(v1) == 1);
    assert(strcmp(str1, (char *) Vector_at( v1, 0 )) == 0);

    Vector_pushBack( v1, str2 );

    assert(Vector_size(v1) == 2);
    assert(Vector_capacity(v1) == 2);
    assert(strcmp(str2, (char *) Vector_at( v1, 1 )) == 0);

    Vector_pushBack( v1, str3 );

    assert(Vector_size(v1) == 3);
    assert(Vector_capacity(v1) == 3);
    assert(strcmp(str3, (char *) Vector_at( v1, 2 )) == 0);

    for(i=0; i < Vector_size(v1); ++i) {
	if(verbose)
	    printf("%4d: %s\n", i, (char *) Vector_at( v1, i ));
    }
    p = Vector_export( v1, 1 );
    assert(strcmp(str2, (char *) p) == 0);
    assert(Vector_at( v1, 1 ) == NULL);
    Vector_setPtr_noCopy(v1, 1, p);
    assert(p == Vector_at( v1, 1 ));


    Vector_resize( v1, 10 );
    assert(Vector_size(v1) == 10);
    assert(Vector_capacity(v1) == 10);
    Vector_setValue( v1, 3, str1 );

    Vector_reserve( v1, 30 );
    assert(Vector_size(v1) == 10);
    assert(Vector_capacity(v1) == 30);

    for(i=0; i < Vector_size(v1); ++i) {
	if(verbose)
	    printf("After reserve: %4d: %s\n", i, (char *) Vector_at(v1,i));
    }

    Vector_clear(v1);
    assert(Vector_size(v1) == 0);
    assert(Vector_capacity(v1) == 30);
    assert(Vector_popBack(v1) == NULL);

    Vector_destroy(v1);     /* destroy empty vector */

    v1 = Vector_construct(eif);
    Vector_reserve(v1,3);
    Vector_pushBack( v1, str1 );
    Vector_pushBack( v1, str2 );
    Vector_pushBack( v1, str3 );

    vtmp1 = Vector_popBack(v1);
    assert(strCompare(&vtmp1, &str3) == 0);
    assert(Vector_size(v1) == 2);

    Vector_assign(v2, v1);
    for(i=0; i < Vector_size(v2); ++i) {
	if(verbose)
	    printf("After assignment: v1[%d]=%s v2[%d]=%s\n",
		   i, (char *) Vector_at(v1,i),
		   i, (char *) Vector_at(v2,i));
    }
    assert(Vector_compare(v1, v2) == 0);

    Vector_append(v1, v2);
    if(verbose) {
	for(i=0; i < Vector_size(v2); ++i) {
	    printf("After append: v1[%d]=%s v2[%d]=%s\n",
		   i, (char *) Vector_at(v1,i),
		   i, (char *) Vector_at(v2,i));
	}
	for(i=Vector_size(v2); i < Vector_size(v1); ++i) {
	    printf("After append: v1[%d]=%s\n",
		   i, (char *) Vector_at(v1,i));
	}
    }
    assert(Vector_size(v1) == 4);
    vtmp1 = Vector_at(v1, 2);
    vtmp2 = Vector_at(v2, 0);
    assert(strCompare(&vtmp1, &vtmp2) == 0);

    vtmp1 = Vector_at(v1, 3);
    vtmp2 = Vector_at(v2, 1);
    assert(strCompare(&vtmp1, &vtmp2) == 0);

    Vector_destroy(v2);
    v2 = Vector_copy(v1);
    assert(Vector_compare(v1, v2) == 0);
    Vector_destroy(v2);

    Vector_clear(v1);
    Vector_pushBack(v1, str3);
    Vector_pushBack(v1, str2);
    Vector_pushBack(v1, str1);
    Vector_pushBack(v1, str2);
    assert(!Vector_isSorted(v1));

    Vector_sort(v1);

    assert(Vector_isSorted(v1));
    assert(Vector_lookup(v1, "test string 0") == 0);
    assert(Vector_lookup(v1, str1) == 0);
    assert(Vector_lookup(v1, str2) == 1);
    assert(Vector_lookup(v1, str3) == 3);
    assert(Vector_lookup(v1, "test string 9") == -1);

    Vector_destroy(v1);     /* destroy full vector */

    ElementInterface_destroy(eif);

    printf("Vector OK\n");
    return 0;
}
